# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os

from lxml import etree as ET

import geni.namespaces as GNS
from .pgmanifest import ManifestSvcLogin, ManifestSvcUser

XPNS = {'g' : GNS.REQUEST.name,
        'u' : GNS.USER.name,
        'v' : "http://geni.bssoftworks.com/rspec/ext/vts/manifest/1",
        's' : "http://geni.bssoftworks.com/rspec/ext/sdn/manifest/1"}

class UnhandledPortTypeError(Exception):
  def __init__ (self, typ):
    super(UnhandledPortTypeError, self).__init__()
    self.typ = typ
  def __str__ (self):
    return "Port type '%s' isn't supported by port builder.  Perhaps you should contribute some code?" % (self.typ)


class GenericPort(object):
  def __init__ (self, typ):
    self.client_id = None
    self.type = typ
    self.cross_sliver = False
    self._name = None
    self._dpname = None

  @classmethod
  def _fromdom (cls, elem):
    p = GenericPort(elem.get("type"))
    p.client_id = elem.get("client_id")
    return p

  def _decomposeClientID (self):
    # Assumes that the client_id is in the format "dp_name:port_name"
    if self.client_id.count(":") == 1:
      (self._dpname, self._name) = self.client_id.split(":")
    # Assumes that the client_id is in the format "dp_name:_x_:port_name"
    elif self.client_id.count(":") == 2: 
      (self._dpname, _, self._name) = self.client_id.split(":")
      self.cross_sliver = True

  @property
  def name (self):
    if not self._name:
      self._decomposeClientID()
    return self._name

  @property
  def dpname (self):
    if not self._dpname:
      self._decomposeClientID()
    return self._dpname


class InternalContainerPort(GenericPort):
  class NoMACAddressError(Exception):
    def __init__ (self, cid):
      super(InternalContainerPort.NoMACAddressError, self).__init__()
      self._cid = cid
    def __str__ (self):
      return "Port with client_id %s does not have MAC address." % (self._cid)

  def __init__ (self):
    super(InternalContainerPort, self).__init__("internal")
    self.remote_client_id = None
    self._macaddress = None
    self._alias = None
    self._remote_dpname = None

  @property
  def macaddress (self):
    if not self._macaddress:
      raise InternalContainerPort.NoMACAddressError(self.client_id)
    else:
      return self._macaddress

  @macaddress.setter
  def macaddress (self, val):
    self._macaddress = val

  @property
  def name (self):
    return self._alias

  @classmethod
  def _fromdom (cls, elem):
    p = InternalContainerPort()
    p.client_id = elem.get("client_id")
    p.remote_client_id = elem.get("remote-clientid")
    p.macaddress = elem.get("mac-address")
    p._alias = elem.get("name")

    return p

  @property
  def remote_dpname (self):
    if not self._remote_dpname:
      if self.remote_client_id.count(":") == 1:
        self._remote_dpname = self.remote_client_id.split(":")[0]
      elif self.remote_client_id.count(":") == 2:
        self._remote_dpname = self.remote_client_id.split(":")[0]
    return self._remote_dpname


class InternalPort(GenericPort):
  def __init__ (self):
    super(InternalPort, self).__init__("internal")
    self.remote_client_id = None
    self._remote_dpname = None
    self._vlan_id = None

  @classmethod
  def _fromdom (cls, elem):
    p = InternalPort()
    p.client_id = elem.get("client_id")
    p.remote_client_id = elem.get("remote-clientid")
    p.vlan_id = int(elem.get("vlan-id", 0))

    return p

  @property
  def remote_dpname (self):
    if not self._remote_dpname:
      if self.remote_client_id.count(":") == 1:
        self._remote_dpname = self.remote_client_id.split(":")[0]
      elif self.remote_client_id.count(":") == 2:
        self._remote_dpname = self.remote_client_id.split(":")[0]
    return self._remote_dpname


class GREPort(GenericPort):
  def __init__ (self):
    super(GREPort, self).__init__("gre")
    self.circuit_plane = None
    self.local_endpoint = None
    self.remote_endpoint = None

  @classmethod
  def _fromdom (cls, elem):
    p = GREPort()
    p.client_id = elem.get("client_id")
    endpe = elem.xpath("v:endpoint", namespaces=XPNS)[0]
    p.circuit_plane = endpe.get("circuit-plane")
    p.local_endpoint = endpe.get("local")
    p.remote_endpoint = endpe.get("remote")
    return p


class PGLocalPort(GenericPort):
  def __init__ (self):
    super(PGLocalPort, self).__init__("pg-local")
    self.shared_vlan = None

  @classmethod
  def _fromdom (cls, elem):
    p = PGLocalPort()
    p.client_id = elem.get("client_id")
    p.shared_vlan = elem.get("shared-lan")
    return p


class VFPort(GenericPort):
  def __init__ (self):
    super(VFPort, self).__init__("vf")
    self.remote_client_id = None

  @classmethod
  def _fromdom (cls, elem):
    p = cls()
    p.client_id = elem.get("client_id")
    p.remote_client_id = elem.get("remote-clientid")
    return p


class ManifestMount(object):
  def __init__ (self):
    self.type = None
    self.name = None
    self.path = None
    self.volid = None

  @classmethod
  def _fromdom (cls, elem):
    m = ManifestMount()
    m.type = elem.get("type")
    m.volid = elem.get("vol-id")
    m.name = elem.get("name")
    m.path = elem.get("path")
    return m


class ManifestContainer(object):
  def __init__ (self):
    self.client_id = None
    self.image = None
    self.sliver_id = None
    self.logins = []
    self.users = []
    self.ports = []
    self.mounts = []

  def __getitem__ (self, key):
    for port in self.ports:
      if port.name == key:
        return port

  @property
  def name (self):
    return self.client_id

  @classmethod
  def _fromdom (cls, elem):
    c = ManifestContainer()
    c.client_id = elem.get("client_id")
    c.image = elem.get("image")
    c.sliver_id = elem.get("sliver_id")

    logins = elem.xpath('g:services/g:login', namespaces = XPNS)
    for lelem in logins:
      l = ManifestSvcLogin._fromdom(lelem)
      c.logins.append(l)

    users = elem.xpath('g:services/u:services_user', namespaces = XPNS)
    for uelem in users:
      u = ManifestSvcUser._fromdom(uelem)
      c.users.append(u)

    ports = elem.xpath('v:port', namespaces = XPNS)
    for cport in ports:
      p = Manifest._buildPort(cport, True)
      c.ports.append(p)

    mounts = elem.xpath('v:mount', namespaces = XPNS)
    for melem in mounts:
      m = ManifestMount._fromdom(melem)
      c.mounts.append(m)

    return c

class ManifestFunction(object):
  def __init__ (self, client_id):
    self.client_id = client_id

  @classmethod
  def _fromdom (cls, elem):
    typ = elem.get("type")
    if typ == "sslvpn":
      f = SSLVPNFunction._fromdom(elem)
      return f


class ManifestDatapath(object):
  def __init__ (self):
    self.client_id = None
    self.image = None
    self.sliver_id = None
    self.ports = []
    self.mirror = None

  def __getitem__ (self, key):
    for port in self.ports:
      if port.name == key:
        return port

  @classmethod
  def _fromdom (cls, elem):
    # TODO: Add ports later
    dp = ManifestDatapath()
    dp.client_id = elem.get("client_id")
    dp.image = elem.get("image")
    dp.sliver_id = elem.get("sliver_id")

    mirror = elem.xpath('v:mirror', namespaces = XPNS)
    if mirror:
      dp.mirror = mirror[0].get("target")

    stp = elem.xpath('v:stp', namespaces = XPNS)
    if stp:
      dp.stp_mode = stp[0].get("type")

    ports = elem.xpath('v:port', namespaces = XPNS)
    for port in ports:
      p = Manifest._buildPort(port)
      dp.ports.append(p)
    return dp


class SSLVPNFunction(ManifestFunction):
  def __init__ (self, client_id):
    super(SSLVPNFunction, self).__init__(client_id)
    self.tp_port = None
    self.local_ip = None
    self.key = None
    self.note = None

  @classmethod
  def _fromdom (cls,elem):
    vpn = SSLVPNFunction(elem.get('client_id'))
    vpn.tp_port = elem.get('tp-port')
    vpn.local_ip = elem.get('local-ip')
    vpn.key = elem.text
    vpn.note = elem.get('note')
    return vpn

class Manifest(object):
  """Wrapper object for GENI XML manifest rspec, providing a pythonic API to the contained data"""

  def __init__ (self, path = None, xml = None):
    if path:
      self._xml = open(path, "r").read()
    elif xml:
      self._xml = xml
    self._root = ET.fromstring(self._xml)
    self._pid = os.getpid()
    self._info = {}

  def _populate_info (self):
    ielems = self._root.xpath('v:info', namespaces = XPNS)
    if ielems:
      self._info["host"] = ielems[0].get("host")
      self._info["slice"] = ielems[0].get("slice")
      self._info["client-topo-name"] = ielems[0].get("client-topo-name")

  def __getitem__ (self, key):
    return self.findTarget(key)

  @property
  def root (self):
    if os.getpid() != self._pid:
      self._root = ET.fromstring(self._xml)
      self._pid = os.getpid()
    return self._root

  @property
  def text (self):
    """String representation of original XML content, with added whitespace for easier reading"""
    return ET.tostring(self.root, pretty_print=True)

  @property
  def pg_circuits (self):
    """Iterator for allocated circuit names on the local PG circuit plane (as strings)."""
    elems = self._root.xpath("v:datapath/v:port[@shared-lan]", namespaces = XPNS)
    for elem in elems:
      yield elem.get("shared-lan")

  local_circuits = pg_circuits

  @property
  def ports (self):
    """Iterator for all datapath and container ports as subclasses of :py:class:`GenericPort` objects."""
    for dp in self.datapaths:
      for p in dp.ports:
        yield p
    for c in self.containers:
      for p in c.ports:
        yield p

  @property
  def containers (self):
    """Iterator over all allocated containers as :py:class:`ManifestContainer` objects."""
    elems = self._root.xpath("v:container", namespaces = XPNS)
    for elem in elems:
      yield ManifestContainer._fromdom(elem)

  @property
  def functions (self):
    """Iterator over all allocated functions as :py:class:`ManifestFunction` objects."""
    elems = self._root.xpath("v:functions/v:function", namespaces = XPNS)
    for elem in elems:
      yield ManifestFunction._fromdom(elem)

  @property
  def datapaths (self):
    """Iterator over all allocated datapaths as :py:class:`ManifestDatapath` objects."""
    elems = self._root.xpath("v:datapath", namespaces = XPNS)
    for elem in elems:
      yield ManifestDatapath._fromdom(elem)

  @property
  def host (self):
    if not self._info:
      self._populate_info()
    return self._info["host"]

  @property
  def slicename (self):
    if not self._info:
      self._populate_info()
    return self._info["slice"]

  @property
  def toponame (self):
    if not self._info:
      self._populate_info()
    return self._info["client-topo-name"]

  def findTarget (self, client_id):
    """Get the container or datapath representing the given `client_id` in the manifest.

    Args:
      client_id (str): Requested client ID of the object you want to find

    Returns:
      :py:class:`ManifestDatapath`, :py:class:`ManifestContainer`, or `None`
    """
      
    dpelems = self._root.xpath("v:datapath[@client_id='%s']" % (client_id), namespaces = XPNS)
    if dpelems:
      return ManifestDatapath._fromdom(dpelems[0])

    ctelems = self._root.xpath("v:container[@client_id='%s']" % (client_id), namespaces = XPNS)
    if ctelems:
      return ManifestContainer._fromdom(ctelems[0])

  def findPort (self, client_id):
    """Get the datapath port object representing the given `client_id`.

    Args:
      client_id (str): client_id of the port you want to find

    Returns:
      :py:class:`GenericPort` or `None`
    """
    pelems = self._root.xpath("v:datapath/v:port[@client_id='%s']" % (client_id), namespaces = XPNS)
    if pelems:
      return Manifest._buildPort(pelems[0])

  @staticmethod
  def _buildPort (elem, container = False):
    t = elem.get("type")
    if t == "gre":
      return GREPort._fromdom(elem)
    elif t == "pg-local":
      return PGLocalPort._fromdom(elem)
    elif t == "vf-port":
      return VFPort._fromdom(elem)
    elif t == "internal":
      if container:
        return InternalContainerPort._fromdom(elem)
      else:
        return InternalPort._fromdom(elem)
    raise UnhandledPortTypeError(t)

  def write (self, path):
    """
.. deprecated:: 0.4
    Use :py:meth:`geni.rspec.vtsmanifest.Manifest.writeXML` instead."""

    import geni.warnings as GW
    import warnings
    warnings.warn("The Manifest.write() method is deprecated, please use Manifest.writeXML() instead",
                  GW.GENILibDeprecationWarning, 2)
    self.writeXML(path)

  def writeXML (self, path):
    """Write the XML representation of this manifest to the supplied path.

    Args:
      path (str): Path to output file
    """
    f = open(path, "w+")
    f.write(ET.tostring(self.root, pretty_print=True))
    f.close()




# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os
import os.path
import datetime

import lxml.etree as ET

class SlicecredProxy(object):
  def __init__ (self, context):
    self._context = context

  def __getitem__ (self, name):
    return self._context._getSliceCred(name)

  def iteritems (self):
    return iter(self._context._slicecred_paths.items())

  def iterkeys (self):
    return iter(self._context._slicecred_paths.keys())

  def __iter__ (self):
    for x in self._context._slicecred_paths:
      yield x

class SliceCredInfo(object):
  class CredentialExpiredError(Exception):
    def __init__ (self, name, expires):
      super(SliceCredInfo.CredentialExpiredError, self).__init__()
      self.expires = expires
      self.sname = name
    def __str__ (self):
      return "Credential for slice %s expired on %s" % (self.sname, self.expires)

  def __init__ (self, context, slicename):
    self.slicename = slicename
    self.context = context
    self._path = None
    self.expires = None
    self.urn = None
    self.type = None
    self.version = None
    self._build()

  def _build (self):
    self._path = "%s/%s-%s-%s-scred.xml" % (self.context.datadir, self.context.cf.name,
                                            self.context.project, self.slicename)
    if not os.path.exists(self._path):
      self._downloadCredential()
    else:
      self._parseInfo()

  def _downloadCredential (self):
    cred = self.context.cf.getSliceCredentials(self.context, self.slicename)

    f = open(self._path, "w+")
    f.write(cred)
    f.close()
    self._parseInfo()

  def _parseInfo (self):
    r = ET.parse(self._path)

    # Expiration
    expstr = r.find("credential/expires").text
    if expstr[-1] == 'Z':
      expstr = expstr[:-1]
    self.expires = datetime.datetime.strptime(expstr, "%Y-%m-%dT%H:%M:%S")

    # URN
    self.urn = r.find("credential/target_urn").text

    # Type
    tstr = r.find("credential/type").text.strip()
    if tstr == "privilege":
      self.type = "geni_sfa"
      self.version = 3  # We hope
    elif tstr == "abac":
      self.type = "abac"
      self.version = 1

  @property
  def path (self):
    checktime = datetime.datetime.now() + datetime.timedelta(days=3)
    if self.expires < checktime:
      # We expire in the next 6 days
      # TODO: Log something
      self._downloadCredential()
      if self.expires < datetime.datetime.now():
        raise SliceCredInfo.CredentialExpiredError(self.slicename, self.expires)
    return self._path

  @property
  def cred_api3 (self):
    scd = {"geni_type" : "geni_sfa", "geni_version" : 3}
    scd["geni_value"] = open(self.path, "r").read()
    return scd


class Context(object):
  DEFAULT_DIR = os.path.expanduser("~/.bssw/geni")

  class UserCredExpiredError(Exception):
    def __init__ (self, expires):
      super(Context.UserCredExpiredError, self).__init__()
      self.expires = expires
    def __str__ (self):
      return "User Credential expired on %s" % (self.expires)

  def __init__ (self):
    self._data_dir = None
    self._nick_cache_path = None
    self._users = set()
    self._cf = None
    self._usercred_info = None  # (path, expires, urn, type, version)
    self._slicecreds = {}
    self.debug = False
    self.uname = None
    self.path = None

#  def save (self):
#    import geni._coreutil as GCU

#    obj = {}
#    obj["framework"] = context.cf.name
#    obj["cert-path"] = context.cf.cert_path
#    obj["key-path"] = context._key

  @property
  def userurn (self):
    return self._cf.userurn

  def _getSliceCred (self, sname):
    info = self.getSliceInfo(sname)
    return info.path

  def _getCredInfo (self, path):
    r = ET.parse(path)
    expstr = r.find("credential/expires").text
    if expstr[-1] == 'Z':
      expstr = expstr[:-1]
    exp = datetime.datetime.strptime(expstr, "%Y-%m-%dT%H:%M:%S")
    urn = r.find("credential/owner_urn").text

    # Type
    tstr = r.find("credential/type").text.strip()
    if tstr == "privilege":
      typ = "geni_sfa"
      version = 3  # We hope
    elif tstr == "abac":
      typ = "abac"
      version = 1

    return (exp, urn, typ, version)

  @property
  def _chargs (self):
    ucinfo = self._ucred_info
    ucd = {"geni_type" : ucinfo[3], "geni_version" : ucinfo[4]}
    ucd["geni_value"] = open(ucinfo[0], "r").read()
    return (False, self.cf.cert, self.cf.key, [ucd])

  @property
  def ucred_api3 (self):
    ucinfo = self._ucred_info
    ucd = {"geni_type" : ucinfo[3], "geni_version" : ucinfo[4]}
    ucd["geni_value"] = open(ucinfo[0], "r").read()
    return ucd

  @property
  def ucred_pg (self):
    return open(self._ucred_info[0], "r").read()

  @property
  def project (self):
    return self.cf.project

  @project.setter
  def project (self, val):
    self.cf.project = val

  @property
  def project_urn (self):
    return self.cf.projecturn

  @property
  def cf (self):
    return self._cf

  @cf.setter
  def cf (self, value):
    # TODO: Calllback into framework here?  Maybe addressed with ISSUE-2
    # Maybe declare writing the _cf more than once as Unreasonable(tm)?
    self._cf = value
    self._usercred_info = None

  @property
  def nickCache (self):
    if self._nick_cache_path is None:
      cachepath = os.path.normpath("%s/nickcache.json" % (self.datadir))
      self._nick_cache_path = cachepath
    return self._nick_cache_path

  @property
  def datadir (self):
    if self._data_dir is None:
      if not os.path.exists(Context.DEFAULT_DIR):
        os.makedirs(Context.DEFAULT_DIR)
      self._data_dir = Context.DEFAULT_DIR
    return self._data_dir

  @datadir.setter
  def datadir (self, val):
    nval = os.path.expanduser(os.path.normpath(val))
    if not os.path.exists(nval):
      os.makedirs(nval)
    self._data_dir = nval

### TODO: User credentials need to belong to Users, or fix up this profile nonsense
  @property
  def _ucred_info (self):
    if (self._usercred_info is None) or (self._usercred_info[1] < datetime.datetime.now()):
      ucpath = "%s/%s-%s-usercred.xml" % (self.datadir, self.cf.name, self.uname)
      if not os.path.exists(ucpath):
        cred = self.cf.getUserCredentials(self.userurn)

        f = open(ucpath, "w+")
        f.write(cred)
        f.close()

      (expires, urn, typ, version) = self._getCredInfo(ucpath)
      self._usercred_info = (ucpath, expires, urn, typ, version)

    if self._usercred_info[1] < datetime.datetime.now():
      cred = self.cf.getUserCredentials(self.userurn)
      f = open(ucpath, "w+")
      f.write(cred)
      f.close()
      (expires, urn, typ, version) = self._getCredInfo(ucpath)
      self._usercred_info = (ucpath, expires, urn, typ, version)

    return self._usercred_info

  @property
  def usercred_path (self):
    # If you only need a user cred, something that works in the next 5 minutes is fine.  If you
    # are doing something more long term then you need a slice credential anyhow, whose
    # expiration will stop you as it should not outlast the user credential (and if it does,
    # some clearinghouse has decided that is allowed).
    checktime = datetime.datetime.now() + datetime.timedelta(minutes=5)
    if self._ucred_info[1] < checktime:
      # Delete the user cred and hope you already renewed it
      try:
        os.remove(self._ucred_info[0])
        self._usercred_info = None
      except OSError:
        # Windows won't let us remove open files
        # TODO: A place for some debug logging
        pass

    if self._ucred_info[1] < datetime.datetime.now():
      raise Context.UserCredExpiredError(self._ucred_info[1])

    return self._ucred_info[0]

  def addUser (self, user):
    self._users.add(user)
    # The first time we call this, it's from context loading, we hope
    # So, the first user is us, and not wacky other people
    # TODO: This is still stupid, and we need to separate framework accounts from resource accounts
    if not self.uname:
      self.uname = user.name

  @property
  def slicecreds (self):
    return SlicecredProxy(self)

  def getSliceInfo (self, sname, project = None):
    if not project:
      project = self.project
    if not self._slicecreds.has_key("%s-%s" % (project, sname)):
      scinfo = SliceCredInfo(self, sname)
      self._slicecreds["%s-%s" % (project, sname)] = scinfo
    return self._slicecreds["%s-%s" % (project, sname)]



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Convenience library to load all extensions supported by Emulab-based
aggregates. In most cases, you will not need to load these extension libraries
individually, just load this one."""

from __future__ import absolute_import

from ..igext import *
from .emuext import *
from .userdata import *

from . import pnext



# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from setuptools import setup, find_packages

import os
import os.path
import platform

requires = ["cryptography",
            "ipaddress",
            "lxml",
            "requests",
            "wrapt"]

# If you are on linux, and don't have ca-certs, we can do an awful thing and it will still work
if os.name == "posix" and os.uname()[0] == "Linux":
  if not os.path.exists("/etc/ssl/certs/ca-certificates.crt"):
    import ssl
    ssl._create_default_https_context = ssl._create_unverified_context

pkgs = find_packages()
pkgs.append('')

setup(name = 'geni-lib',
      version = '0.9.7.10',
      author = 'Nick Bastin',
      author_email = 'nick@bssoftworks.com',
      description = 'Library and tools for working with research testbed resources that support ' \
                    'the GENI AM API, including the NSF GENI Testbed (www.geni.net) and Cloudlab (cloudlab.us).',
      long_description = open("README.rst", "r").read(),
      packages = pkgs,
      package_dir = {'' : 'tools/ipython', 'geni' : 'geni', 'ccloud' : 'ccloud'},
      pymodules = ['genish'],
      scripts = ['tools/buildcontext/context-from-bundle',
                 'tools/buildcontext/build-context',
                 'tools/shell/genish'],
      url = 'https://bitbucket.org/barnstorm/geni-lib',
      install_requires = requires,
      classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        ]
      )



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from ..pg import RawPC, Execute, Request
from ..igext import XenVM
from .epcexc import InvalidRole
from .pndefs import EPCROLES, PNDEFS
from .emuext import ProgramAgent

#
# Defaults/constants used below
#
class _EPCDEFS(object):
    OEPC_STARTSCRIPT = "/opt/OpenEPC/bin/start_epc.sh"
    SUDOBIN = "/usr/bin/sudo"
    VM_RAM = 1024
    VALID_ROLES = (EPCROLES.ANY, EPCROLES.ENABLERS, EPCROLES.PGW, 
                   EPCROLES.SGW_MME_SGSN, EPCROLES.CLIENT, EPCROLES.ENODEB)

#
# Factory method for returning VM or raw node based on global setting.
# Settings class also allows users to set global image and hardware type
# for EPC nodes.
#
class EPCNodeFactorySettings(object):
    use_vm_nodes = False
    hardware_type = None
    disk_image = PNDEFS.DEF_BINOEPC_IMG
    do_sync_start = False

def mkepcnode(*args, **kwargs):
    node = None
    request = None
    if "request" in kwargs:
        request = kwargs["request"]
        del kwargs["request"]
    if EPCNodeFactorySettings.use_vm_nodes:
        if request:
            node = request.EPCVMNode(*args, **kwargs)
        else:
            node = EPCVMNode(*args, **kwargs)
    else:
        if request:
            node = request.EPCNode(*args, **kwargs)
        else:
            node = EPCNode(*args, **kwargs)
    if EPCNodeFactorySettings.hardware_type:
        node.hardware_type = EPCNodeFactorySettings.hardware_type
    if EPCNodeFactorySettings.disk_image:
        node.disk_image = EPCNodeFactorySettings.disk_image
    if EPCNodeFactorySettings.do_sync_start:
        node.syncstart = EPCNodeFactorySettings.do_sync_start
    return node

#
# Base EPC mixin node class.  This is the bulk of the implementation for
# both physical and VM node classes.
#
class _EPCBaseNode(object):
    def __init__ (self, client_id, role, hname = None, component_id = None,
                  prehook = None, posthook = None):
        if not role in _EPCDEFS.VALID_ROLES:
            raise InvalidRole(role)
        super(_EPCBaseNode, self).__init__(client_id, component_id = component_id)
        self.role = role
        self.hname = hname
        self.startscript = _EPCDEFS.OEPC_STARTSCRIPT
        self.prehook = prehook
        self.posthook = posthook
        self.syncstart = False

    def _write(self, root):
        startcmd = "%s %s -r %s" % (_EPCDEFS.SUDOBIN, self.startscript, 
                                    self.role)
        if self.hname:
            startcmd += " -h %s" % self.hname
        if self.prehook:
            startcmd += " -P %s" % self.prehook
        if self.posthook:
            startcmd += " -T %s" % self.posthook
        if self.syncstart:
            self.addService(ProgramAgent(self.client_id + '_epc0', startcmd, None, True))
        else:
            self.addService(Execute(shell="csh", command=startcmd))
        return super(_EPCBaseNode, self)._write(root)


#
# Physical EPC node class
#
class EPCNode(_EPCBaseNode,RawPC):
    pass

Request.EXTENSIONS.append(("EPCNode", EPCNode))


#
# VM EPC node class
#
class EPCVMNode(_EPCBaseNode,XenVM):
    def __init__ (self, client_id, role, hname = None, component_id = None,
                  prehook = None, posthook = None):
        super(EPCVMNode, self).__init__(client_id, role, hname, component_id, 
                                        prehook, posthook)
        self.exclusive = True # NEVER run on a shared host.
        self.disk = 0 # No thin provisioning when set!
        self.ram = _EPCDEFS.VM_RAM

Request.EXTENSIONS.append(("EPCVMNode", EPCVMNode))



# Copyright (c) 2014-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import inspect
import sys

from .core import AM

class Transit(AM):
  def __init__ (self, name, amtype, cmid, url):
    super(Transit, self).__init__(name, url, "amapiv2", amtype, cmid)


AL2S = Transit("i2-al2s", "oess", "urn:publicid:IDN+al2s.internet2.edu+authority+am",
               "https://geni-al2s.net.internet2.edu:3626/foam/gapi/2")

ION = Transit("i2-ion", "pg", "urn:publicid:IDN+ion.internet2.edu+authority+am",
              "http://geni-am.net.internet2.edu:12346")

MAX = Transit("dcn-max", "pg", "urn:publicid:IDN+dragon.maxgigapop.net+authority+am",
              "http://max-myplc.dragon.maxgigapop.net:12346")

Utah = Transit("utah-stitch", "pg", "urn:publicid:IDN+stitch.geniracks.net+authority+cm",
               "https://stitch.geniracks.net:12369/protogeni/xmlrpc/am")

def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      result[obj.name] = obj
  return result



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from ..pg import Link, Request

#
# Class to hide the idiosynchrasies of PhantomNet RF links.
#
class RFLink(Link):
    def __init__(self, name):
        super(RFLink, self).__init__(name)
        self.bandwidth = 500
  
    def _write(self, root):
        lnk = super(RFLink, self)._write(root)
        lnk.attrib["protocol"] = "P2PLTE"
        return lnk

Request.EXTENSIONS.append(("RFLink", RFLink))



# Copyright (c) 2013-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import inspect
import sys

from .core import AM

class OGCompute(AM):
  def __init__ (self, name, host, cmid = None, url = None):
    if url is None:
      url = "https://%s:5002" % (host)
    super(OGCompute, self).__init__(name, url, "amapiv2", "opengeni", cmid)

GPO_OG = OGCompute("gpo-og", "bbn-cam-ctrl-1.gpolab.bbn.com", "urn:publicid:IDN+bbn-cam-ctrl-1.gpolab.bbn.com+authority+am")
CLEMSON_OG = OGCompute("clemson-og", "clemson-clemson-control-1.clemson.edu",
                       "urn:publicid:IDN+clemson-clemson-control-1.clemson.edu+authority+am")
UKL_OG = OGCompute("ukl-og", "glab077.e4.ukl.german-lab.de", "urn:publicid:IDN+glab077.e4.ukl.german-lab.de:gcf+authority+am")

def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      result[obj.name] = obj
  return result



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import geni.urn
import geni.aggregate.protogeni

#
# Static set of EPC roles
#
class EPCROLES(object):
    ANY          = "any"
    ENABLERS     = "epc-enablers"
    PGW          = "pgw"
    SGW_MME_SGSN = "sgw-mme-sgsn"
    CLIENT       = "epc-client"
    ENODEB       = "enodeb"

#
# Static set of EPC lans (recognized by PhantomNet setup code)
#
class EPCLANS(object):
    MGMT   = "mgmt"
    NET_A  = "net-a"
    NET_B  = "net-b"
    NET_C  = "net-c"
    NET_D  = "net-d"
    AN_LTE = "an-lte"

#
# Other global constants
#
class PNDEFS(object):
    PNET_AM = geni.aggregate.protogeni.UTAH_PG
    DEF_BINOEPC_IMG = geni.urn.Image(PNET_AM, "PhantomNet:UBUNTU12-64-BINOEPC")



# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import, print_function

import datetime
import json
import multiprocessing as MP
import os
import os.path
import shutil
import subprocess
import tempfile
import time
import traceback as tb
import zipfile

from .aggregate.apis import ListResourcesError, DeleteSliverError

def _getdefault (obj, attr, default):
  if hasattr(obj, attr):
    return obj[attr]
  return default

def checkavailrawpc (context, am):
  """Returns a list of node objects representing available raw PCs at the
given aggregate."""

  avail = []
  ad = am.listresources(context)
  for node in ad.nodes:
    if node.exclusive and node.available:
      if "raw-pc" in node.sliver_types:
        avail.append(node)
  return avail


def _corelogininfo (manifest):
  from .rspec.vtsmanifest import Manifest as VTSM
  from .rspec.pgmanifest import Manifest as PGM

  linfo = []
  if isinstance(manifest, PGM):
    for node in manifest.nodes:
      linfo.extend([(node.client_id, x.username, x.hostname, x.port) for x in node.logins])
  elif isinstance(manifest, VTSM):
    for container in manifest.containers:
      linfo.extend([(container.client_id, x.username, x.hostname, x.port) for x in container.logins])
  return linfo


def printlogininfo (context = None, am = None, slice = None, manifest = None):
  """Prints out host login info in the format:
::
  [client_id][username] hostname:port

If a manifest object is provided the information will be mined from this data,
otherwise you must supply a context, slice, and am and a manifest will be
requested from the given aggregate."""

  if not manifest:
    manifest = am.listresources(context, slice)

  info = _corelogininfo(manifest)
  for line in info:
    print("[%s][%s] %s: %d" % (line[0], line[1], line[2], line[3]))


# You can't put very much information in a queue before you hang your OS
# trying to write to the pipe, so we only write the paths and then load
# them again on the backside
def _mp_get_manifest (context, site, slc, q):
  try:
    # Don't use geni.tempfile here - we don't want them deleted when the child process ends
    # TODO: tempfiles should get deleted when the parent process picks them back up
    mf = site.listresources(context, slc)
    tf = tempfile.NamedTemporaryFile(delete=False)
    tf.write(mf.text)
    path = tf.name
    tf.close()
    q.put((site.name, slc, path))
  except ListResourcesError:
    q.put((site.name, slc, None))
  except Exception:
    tb.print_exc()
    q.put((site.name, slc, None))

def getManifests (context, ams, slices):
  """Returns a two-level dictionary of the form:
::
  {slice_name : { site_object : manifest_object, ... }, ...}

Containing the manifests for all provided slices at all the provided
sites.  Requests are made in parallel and the function blocks until the
slowest site returns (or times out)."""

  sitemap = {}
  for am in ams:
    sitemap[am.name] = am

  q = MP.Queue()
  for site in ams:
    for slc in slices:
      p = MP.Process(target=_mp_get_manifest, args=(context, site, slc, q))
      p.start()

  while MP.active_children():
    time.sleep(0.5)

  d = {}
  while not q.empty():
    (site,slc,mpath) = q.get()

    if mpath:
      am = sitemap[site]
      data = open(mpath).read()
      mf = am.amtype.parseManifest(data)
      d.setdefault(slc, {})[sitemap[site]] = mf

  return d


def _mp_get_advertisement (context, site, q):
  try:
    ad = site.listresources(context)
    q.put((site.name, ad))
  except Exception:
    q.put((site.name, None))

def getAdvertisements (context, ams):
  """Returns a dictionary of the form:
::
  { site_object : advertisement_object, ...}

Containing the advertisements for all the requested aggregates.  Requests
are made in parallel and the function blocks until the slowest site
returns (or times out).

.. warning::
  Particularly large advertisements may break the shared memory queue
  used by this function."""


  q = MP.Queue()
  for site in ams:
    p = MP.Process(target=_mp_get_advertisement, args=(context, site, q))
    p.start()

  while MP.active_children():
    time.sleep(0.5)

  d = {}
  while not q.empty():
    (site,ad) = q.get()
    d[site] = ad

  return d


def deleteSliverExists(am, context, slice):
  """Attempts to delete all slivers for the given slice at the given AM, suppressing all returned errors."""
  try:
    am.deletesliver(context, slice)
  except DeleteSliverError:
    pass

def _buildaddot(ad, drop_nodes = None):
  """Constructs a dotfile of a topology described by an advertisement rspec.  Only works on very basic GENIv3 advertisements,
  and probably has lots of broken edge cases."""
  # pylint: disable=too-many-branches

  if not drop_nodes:
    drop_nodes = []

  dot_data = []
  dda = dot_data.append # Save a lot of typing

  dda("graph {")

  for node in ad.nodes:
    if node.name in drop_nodes:
      continue

    if node.available:
      dda("\"%s\"" % (node.name))
    else:
      dda("\"%s\" [style=dashed]" % (node.name))

  for link in ad.links:
    if not len(link.interface_refs) == 2:
      print("Link with more than 2 interfaces:")
      print(link.text)

    name_1 = link.interface_refs[0].split(":")[-2].split("+")[-1]
    name_2 = link.interface_refs[1].split(":")[-2].split("+")[-1]

    if name_1 in drop_nodes or name_2 in drop_nodes:
      continue

    dda("\"%s\" -- \"%s\"" % (name_1, name_2))

  dda("}")

  return "\n".join(dot_data)


def builddot (manifests):
  """Constructs a dotfile of the topology described in the passed in manifest list and returns it as a string."""
  # pylint: disable=too-many-branches

  from .rspec import vtsmanifest as VTSM
  from .rspec.pgmanifest import Manifest as PGM

  dot_data = []
  dda = dot_data.append # Save a lot of typing

  dda("digraph {")

  for manifest in manifests:
    if isinstance(manifest, PGM):
      intf_map = {}
      for node in manifest.nodes:
        dda("\"%s\" [label = \"%s\"]" % (node.sliver_id, node.name))
        for interface in node.interfaces:
          intf_map[interface.sliver_id] = (node, interface)

      for link in manifest.links:
        label = link.client_id
        name = link.client_id

        if link.vlan:
          label = "VLAN\n%s" % (link.vlan)
          name = link.vlan

        dda("\"%s\" [label=\"%s\",shape=doublecircle,fontsize=11.0]" % (name, label))

        for ref in link.interface_refs:
          dda("\"%s\" -> \"%s\" [taillabel=\"%s\"]" % (
            intf_map[ref][0].sliver_id, name,
            intf_map[ref][1].component_id.split(":")[-1]))
          dda("\"%s\" -> \"%s\"" % (name, intf_map[ref][0].sliver_id))


    elif isinstance(manifest, VTSM.Manifest):
      for dp in manifest.datapaths:
        dda("\"%s\" [shape=rectangle];" % (dp.client_id))

      for ctr in manifest.containers:
        dda("\"%s\" [shape=oval];" % (ctr.client_id))

      dda("subgraph cluster_vf {")
      dda("label = \"SSL VPNs\";")
      dda("rank = same;")
      for vf in manifest.functions:
        if isinstance(vf, VTSM.SSLVPNFunction):
          dda("\"%s\" [label=\"%s\",shape=hexagon];" % (vf.client_id, vf.note))
      dda("}")

      # TODO: We need to actually go through datapaths and such, but we can approximate for now
      for port in manifest.ports:
        if isinstance(port, VTSM.GREPort):
          pass
        elif isinstance(port, VTSM.PGLocalPort):
          dda("\"%s\" -> \"%s\" [taillabel=\"%s\"]" % (port.dpname, port.shared_vlan,
                                                       port.name))
          dda("\"%s\" -> \"%s\"" % (port.shared_vlan, port.dpname))
        elif isinstance(port, VTSM.InternalPort):
          dp = manifest.findTarget(port.dpname)
          if dp.mirror == port.client_id:
            continue # The other side will handle it, oddly
          # TODO: Handle mirroring into another datapath
          dda("\"%s\" -> \"%s\" [taillabel=\"%s\"]" % (port.dpname, port.remote_dpname,
                                                       port.name))
        elif isinstance(port, VTSM.InternalContainerPort):
          # Check to see if the other side is a mirror into us
          dp = manifest.findTarget(port.remote_dpname)
          if isinstance(dp, VTSM.ManifestDatapath):
            if port.remote_client_id == dp.mirror:
              remote_port_name = port.remote_client_id.split(":")[-1]
              dda("\"%s\" -> \"%s\" [headlabel=\"%s\",taillabel=\"%s\",style=dashed]" % (
                  port.remote_dpname, port.dpname, port.name, remote_port_name))
              continue

          # No mirror, draw as normal
          dda("\"%s\" -> \"%s\" [taillabel=\"%s\"]" % (port.dpname, port.remote_dpname,
                                                       port.name))
        elif isinstance(port, VTSM.VFPort):
          dda("\"%s\" -> \"%s\"" % (port.dpname, port.remote_client_id))
          dda("\"%s\" -> \"%s\"" % (port.remote_client_id, port.dpname))

        elif isinstance(port, VTSM.GenericPort):
          pass
        else:
          continue ### TODO: Unsupported Port Type


  dda("}")

  return "\n".join(dot_data)


class APIEncoder(json.JSONEncoder):
  def default (self, obj): # pylint: disable=E0202
    if hasattr(obj, "__json__"):
      return obj.__json__()
    elif isinstance(obj, set):
      return list(obj)
    return json.JSONEncoder.default(self, obj)


def loadAggregates (path = None):
  from .aggregate.spec import AMSpec
  from . import _coreutil as GCU

  if not path:
    path = GCU.getDefaultAggregatePath()

  ammap = {}
  try:
    obj = json.loads(open(path, "r").read())
    for aminfo in obj["specs"]:
      ams = AMSpec._jconstruct(aminfo)
      am = ams.build()
      if am:
        ammap[am.name] = am
  except IOError:
    pass

  return ammap

def updateAggregates (context, ammap):
  from .aggregate.core import loadFromRegistry

  new_map = loadFromRegistry(context)
  for k,v in new_map.items():
    if k not in ammap:
      ammap[k] = v
  saveAggregates(ammap)

def saveAggregates (ammap, path = None):
  from . import _coreutil as GCU

  if not path:
    path = GCU.getDefaultAggregatePath()

  obj = {"specs" : [x._amspec for x in ammap.values() if x._amspec]}
  with open(path, "w+") as f:
    data = json.dumps(obj, cls=APIEncoder)
    f.write(data)


def loadContext (path = None, key_passphrase = None):
  import geni._coreutil as GCU
  from geni.aggregate import FrameworkRegistry
  from geni.aggregate.context import Context
  from geni.aggregate.user import User

  if path is None:
    path = GCU.getDefaultContextPath()
  else:
    path = os.path.expanduser(path)

  obj = json.load(open(path, "r"))

  version = _getdefault(obj, "version", 1)

  if key_passphrase is True:
    import getpass
    key_passphrase = getpass.getpass("Private key passphrase: ")

  if version == 1:
    cf = FrameworkRegistry.get(obj["framework"])()
    cf.cert = obj["cert-path"]
    if key_passphrase:
      cf.setKey(obj["key-path"], key_passphrase)
    else:
      cf.key = obj["key-path"]

    user = User()
    user.name = obj["user-name"]
    user.urn = obj["user-urn"]
    user.addKey(obj["user-pubkeypath"])

    context = Context()
    context.addUser(user)
    context.cf = cf
    context.project = obj["project"]
    context.path = path

  elif version == 2:
    context = Context()

    fobj = obj["framework-info"]
    cf = FrameworkRegistry.get(fobj["type"])()
    cf.cert = fobj["cert-path"]
    if key_passphrase:
      cf.setKey(fobj["key-path"], key_passphrase)
    else:
      cf.key = fobj["key-path"]
    context.cf = cf
    context.project = fobj["project"]
    context.path = path

    ulist = obj["users"]
    for uobj in ulist:
      user = User()
      user.name = uobj["username"]
      user.urn = _getdefault(uobj, "urn", None)
      klist = uobj["keys"]
      for keypath in klist:
        user.addKey(keypath)
      context.addUser(user)

  from cryptography import x509
  from cryptography.hazmat.backends import default_backend
  cert = x509.load_pem_x509_certificate(open(context._cf.cert, "rb").read(), default_backend())
  if cert.not_valid_after < datetime.datetime.now():
    print("***WARNING*** Client SSL certificate supplied in this context is expired")
  return context


def hasDataContext ():
  import geni._coreutil as GCU

  path = GCU.getDefaultContextPath()
  return os.path.exists(path)


class MissingPublicKeyError(Exception):
  def __str__ (self):
    return "Your bundle does not appear to contain an SSH public key.  You must supply a path to one."


class PathNotFoundError(Exception):
  def __init__ (self, path):
    super(PathNotFoundError, self).__init__()
    self._path = path

  def __str__ (self):
    return "The path %s does not exist." % (self._path)

def _find_ssh_keygen ():
  PATHS = ["/usr/bin/ssh-keygen", "/bin/ssh-keygen", "/usr/sbin/ssh-keygen", "/sbin/ssh-keygen"]
  for path in PATHS:
    if os.path.exists(path):
      return path

MAKE_KEYPAIR = (-1, 1)

def buildContextFromBundle (bundle_path, pubkey_path = None, cert_pkey_path = None):
  import geni._coreutil as GCU

  HOME = os.path.expanduser("~")

  # Create the .bssw directories if they don't exist
  DEF_DIR = GCU.getDefaultDir()

  zf = zipfile.ZipFile(os.path.expanduser(bundle_path))

  zip_pubkey_path = None
  if pubkey_path is None or pubkey_path == MAKE_KEYPAIR:
    # search for pubkey-like file in zip
    for fname in zf.namelist():
      if fname.startswith("ssh/public/") and fname.endswith(".pub"):
        zip_pubkey_path = fname
        break

    if not zip_pubkey_path and pubkey_path != MAKE_KEYPAIR:
      raise MissingPublicKeyError()

  # Get URN/Project/username from omni_config
  urn = None
  project = None

  oc = zf.open("omni_config")
  for l in oc.readlines():
    if l.startswith("urn"):
      urn = l.split("=")[1].strip()
    elif l.startswith("default_project"):
      project = l.split("=")[1].strip()

  uname = urn.rsplit("+")[-1]

  # Create .ssh if it doesn't exist
  try:
    os.makedirs("%s/.ssh" % (HOME), 0775)
  except OSError:
    pass

  # If a pubkey wasn't supplied on the command line, we may need to install both keys from the bundle
  # This will catch if creation was requested but failed
  pkpath = pubkey_path
  if not pkpath or pkpath == MAKE_KEYPAIR:
    found_private = False

    if "ssh/private/id_geni_ssh_rsa" in zf.namelist():
      found_private = True
      if not os.path.exists("%s/.ssh/id_geni_ssh_rsa" % (HOME)):
        # If your umask isn't already 0, we can't safely create this file with the right permissions
        with os.fdopen(os.open("%s/.ssh/id_geni_ssh_rsa" % (HOME), os.O_WRONLY | os.O_CREAT, 0o600), "w") as tf:
          tf.write(zf.open("ssh/private/id_geni_ssh_rsa").read())

    if zip_pubkey_path:
      pkpath = "%s/.ssh/%s" % (HOME, zip_pubkey_path[len('ssh/public/'):])
      if not os.path.exists(pkpath):
        with open(pkpath, "w+") as tf:
          tf.write(zf.open(zip_pubkey_path).read())

    # If we don't find a proper keypair, we'll make you one if you asked for it
    # This preserves your old pubkey if it existed in case you want to use that later
    if not found_private and pubkey_path == MAKE_KEYPAIR:
      keygen = _find_ssh_keygen()
      subprocess.call("%s -t rsa -b 2048 -f ~/.ssh/genilib_rsa -N ''" % (keygen), shell = True)
      pkpath = os.path.expanduser("~/.ssh/genilib_rsa.pub")
  else:
    pkpath = os.path.expanduser(pubkey_path)
    if not os.path.exists(pkpath):
      raise PathNotFoundError(pkpath)

  # We write the pem into 'private' space
  zf.extract("geni_cert.pem", DEF_DIR)

  if cert_pkey_path is None:
    ckpath = "%s/geni_cert.pem" % (DEF_DIR)
  else:
    # Use user-provided key path instead of key inside .pem
    ckpath = os.path.expanduser(cert_pkey_path)
    if not os.path.exists(ckpath):
      raise PathNotFoundError(ckpath)

  cdata = {}
  cdata["framework"] = "portal"
  cdata["cert-path"] = "%s/geni_cert.pem" % (DEF_DIR)
  cdata["key-path"] = ckpath
  cdata["user-name"] = uname
  cdata["user-urn"] = urn
  cdata["user-pubkeypath"] = pkpath
  cdata["project"] = project
  json.dump(cdata, open("%s/context.json" % (DEF_DIR), "w+"))


def _buildContext (framework, cert_path, key_path, username, user_urn, pubkey_path, project, path=None):
  import geni._coreutil as GCU

  # Create the .bssw directories if they don't exist
  DEF_DIR = GCU.getDefaultDir()

  new_cert_path = "%s/%s" % (DEF_DIR, os.path.basename(cert_path))
  shutil.copyfile(cert_path, new_cert_path)

  if key_path != cert_path:
    new_key_path = "%s/%s" % (DEF_DIR, os.path.basename(key_path))
    shutil.copyfile(key_path, new_key_path)
  else:
    new_key_path = new_cert_path

  if not path:
    path = "%s/context.json" % (DEF_DIR)

  cdata = {}
  cdata["framework"] = framework
  cdata["cert-path"] = new_cert_path
  cdata["key-path"] = new_key_path
  cdata["user-name"] = username
  cdata["user-urn"] = user_urn
  cdata["user-pubkeypath"] = pubkey_path
  cdata["project"] = project
  json.dump(cdata, open(path, "w+"))



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

#
# Common EPC exception classes
#

class InvalidRole(Exception):
    def __init__(self, role):
        self.role = role
    def __str__(self):
        return "Warning: Invalid EPC role: %s" % role



# Copyright (c) 2015-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os
import os.path
import ssl

import pkg_resources

WIN32_ATTR_HIDDEN = 0x02
VERSION = pkg_resources.require("geni-lib")[0].version

def getDefaultDir ():
  HOME = os.path.expanduser("~")

  if os.name == "posix":
    DEF_DIR = "%s/.bssw/geni/" % (HOME)
    if not os.path.exists(DEF_DIR):
      os.makedirs(DEF_DIR, 0o775)
  elif os.name == "nt":
    DEF_DIR = "%s/bssw/geni/" % (HOME)
    if not os.path.exists(DEF_DIR):
      os.makedirs(DEF_DIR, 0o775)
      import ctypes
      BSSW = "%s/bssw" % (HOME)
      if not ctypes.windll.kernel32.SetFileAttributesW(unicode(BSSW), WIN32_ATTR_HIDDEN):
        raise ctypes.WinError()

  return DEF_DIR

def getDefaultAggregatePath():
  ddir = getDefaultDir()
  return "%s/aggregates.json" % (ddir)

def getOSName ():
  if os.name == "posix":
    return "%s-%s" % (os.uname()[0], os.uname()[4])
  elif os.name == "nt":
    import platform
    return "%s-%s" % (platform.platform(), platform.architecture()[0])
  return "unknown"


def defaultHeaders ():
  d = {"User-Agent" : "GENI-LIB %s (%s)" % (VERSION, getOSName())}
  return d

def getDefaultContextPath ():
  ddir = getDefaultDir()
  return os.path.normpath("%s/context.json" % (ddir))

def disableUrllibWarnings ():
  try:
    import requests.packages.urllib3
    import requests.packages.urllib3.exceptions

    warnings = []
    try:
      warnings.append(requests.packages.urllib3.exceptions.InsecureRequestWarning)
    except AttributeError:
      pass

    try:
      warnings.append(requests.packages.urllib3.exceptions.InsecurePlatformWarning)
    except AttributeError:
      pass

    try:
      warnings.append(requests.packages.urllib3.exceptions.SNIMissingWarning)
    except AttributeError:
      pass

    requests.packages.urllib3.disable_warnings(tuple(warnings))
  except ImportError:
    # This version of requests doesn't have urllib3 in it
    return

def shellImports ():
  imports = {}

  import geni.constants
  import geni.util
  import geni.rspec.pg
  import geni.rspec.vts
  import geni.rspec.igext
  import geni.rspec.egext
  import geni.rspec.igutil
  import geni.aggregate.frameworks
  import geni.aggregate.instageni
  import geni.aggregate.vts
  import geni.aggregate.exogeni
  import geni.aggregate.cloudlab
  import geni.aggregate.transit
  import pprint

  imports["util"] = geni.util
  imports["PG"] = geni.rspec.pg
  imports["VTS"] = geni.rspec.vts
  imports["IGAM"] = geni.aggregate.instageni
  imports["VTSAM"] = geni.aggregate.vts
  imports["EGAM"] = geni.aggregate.exogeni
  imports["CLAM"] = geni.aggregate.cloudlab
  imports["TRANSITAM"] = geni.aggregate.transit
  imports["IGX"] = geni.rspec.igext
  imports["EGX"] = geni.rspec.egext
  imports["IGUtil"] = geni.rspec.igutil
  imports["PP"] = pprint.pprint
  imports["RegM"] = geni.aggregate.frameworks.MemberRegistry
  imports["SLICE_ROLE"] = geni.constants.SLICE_ROLE
  imports["PROJECT_ROLE"] = geni.constants.PROJECT_ROLE
  imports["REQSTATUS"] = geni.constants.REQSTATUS

  return imports


# pylint: disable=wrong-import-position
from requests.adapters import HTTPAdapter
try:
  from requests.packages.urllib3.poolmanager import PoolManager
  class TLSHttpAdapter(HTTPAdapter):
    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
      self.poolmanager = PoolManager(num_pools = connections, maxsize = maxsize,
                                     block = block, ssl_version = ssl.PROTOCOL_TLSv1)
except ImportError:
  TLSHttpAdapter = HTTPAdapter
